package JSPservletPkg;
import java.util.*;
import java.net.*;
import java.io.*;
/**
 * Global cache for Mapper.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
class Cache implements CacheIntf {
  /** Version 1.0.4. Stores country code instead of PageBox array.<p>
   * Stores a range of requestors along with the country code.
   * Validated with mapper.testMisc.</p>
   */
  class Entry {
    /** First requestor address in the range. */
    long address1;
    /** Last requestor address in the range. */
    long address2;
    /** Timestamp. */
    long time;
    /** Country code. */
    String code;
    /**
     * Constructor.
     * @param address1 first requestor address in the range.
     * @param address2 last requestor address in the range.
     * @param code country code.
     */
    Entry(long address1, long address2, String code) {
      this.address1 = address1;
      this.address2 = address2;
      this.code = code;
      time = System.currentTimeMillis();
    }
  }
  Cache(String cacheLocation) {
    filename = cacheLocation;
    File f = new File(cacheLocation);
    if (f.exists()) {
      try {
        ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(
          new FileInputStream(f)));
        cache = (ArrayList)ois.readObject();
        ois.close();
      }
      catch(Exception e) {}
    }
    if (cache == null)
      cache = new ArrayList();
  }
  /** Cache save.*/
  public synchronized boolean save() {
    try {
      ObjectOutputStream bos = new ObjectOutputStream(
        new BufferedOutputStream(new FileOutputStream(filename)));
      bos.writeObject(cache);
      bos.close();
    }
    catch(Exception e) {
      return false;
    }
    return true;
  }
  /**
   * Address range to PageBoxes cache. Created by initialize. Used by search.<br>
   * Entries: class CacheEntry.
   */
  ArrayList cache = null;
  /** Persistent cache path. */
  String filename;
  /** Returns a report string about cache content. */
  public final synchronized String report() {
    Iterator i = cache.iterator();
    String rep = "";
    while(i.hasNext()) {
      Entry ce = (Entry)i.next();
      rep += "<br>\n" + Mapper.formatAddress(ce.address1) + " " +
        Mapper.formatAddress(ce.address2) + " " + ce.code;
    }
    return rep;
  }
  /**
   * Register a new range and PageBox set in cache.<br>
   * Validated with mapper.testMisc.
   * @param address1 first requestor address in the range.
   * @param address2 last requestor address in the range.
   * @param code country code.
   */
  public final synchronized void register(long address1, long address2,
    String code) {
    cache.add(new Entry(address1, address2, code));
  }
  /**
   * Retrieve a country code from cache.<br>
   * Validated with mapper.testMisc.
   * @param address requestor address.
   */
  public final synchronized String retrieve(long address) {
    Iterator i = cache.iterator();
    while(i.hasNext()) {
      Entry ce = (Entry)i.next();
      if (address < ce.address1)
        continue;
      if (address > ce.address2)
        continue;
      ce.time = System.currentTimeMillis();
      return ce.code;
    }
    return null;
  }
  /**
   * Remove a url mapping from cache.<br>
   * Validated with mapper.testMisc.
   * @param address requestor address - any address in the range.
   */
  public final synchronized void remove(long address) {
    Iterator i = cache.iterator();
    while(i.hasNext()) {
      Entry ce = (Entry)i.next();
      if (address < ce.address1)
        continue;
      if (address > ce.address2)
        continue;
      i.remove();
    }
  }
  /**
   * Remove cache entries older than nbdays.
   * @param nbdays maximum life in cache (days).
   */
  public synchronized void clear(int nbdays) {
    if (nbdays <= 0)
      return;
    long last = System.currentTimeMillis() - (nbdays * 24 * 3600000);
    Iterator i = cache.iterator();
    while(i.hasNext()) {
      Entry ce = (Entry)i.next();
      if (ce.time < last)
        i.remove();
    }
  }
}

