package JSPservletPkg;
import java.lang.*;
import java.util.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;
import java.security.*;
import java.net.*;
import javax.servlet.*;
/**
 * Class loader for classes included on ClassPath.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
public class ClassPathLoader extends ClassLoader {
  /** Calling JSPhandler. */
  public JSPhandler handler;
  /** parent class loader. */
  ClassLoader parent;
  /**
   * classPath classes.<ul>
   * <li>Key: class name.</li>
   * <li>Value: class.</li></ul>
   */
  HashMap classes = new HashMap();
  /**
   * Binary warehouse.<ul>
   * <li>Key: class name.</li>
   * <li>Value: byte array (class binary).</li></ul>
   */
  HashMap classBufs = new HashMap();
  /**
   * resources value - maintains a buffer containing the resource and a
   * lastModified time.
   */
  class ResourceEntry {
    /** time in milliseconds since the epoch where the resource was modified */
    long lastModified = -1;
    /** buffer containing the resource */
    byte[] buf = null;
  }
  /**
   * classPath resources.<ul>
   * <li>Key: class name.</li>
   * <li>Value: ResourceEntry.</li></ul>
   */
  HashMap resources = new HashMap();
  /** URI that triggered the creation of the instance. */
  String URI;
  /** Path of URI. Used by getResourceAsStreamFromArchive. */
  String URIRealPath;
  /**
   * Creates classes stored in classPath and stores them in classes.
   * @param  jh calling JSPhandler.
   * @param   URI Request URL up to context path.
   *          Used by ClassPathLoader to handle resources.
   * @param   URIRealPath path without servlet part.
   *          Used in getResourceAsStreamFromArchive.
   */
  public ClassPathLoader(JSPhandler jh, String URI, String URIRealPath) {
    super();
    handler = jh;
    parent = getParent();
    if (parent == null)
      jh.log.logprint(toString() + ".ClassPathLoader null parent");
    this.URI = "" + URI;
    this.URIRealPath = URIRealPath;
    for (int i = 0; i < handler.classPath.length; ++ i)
      loadClassData(handler.classPath[i]);
    defineClasses();
    classBufs.clear();
    classBufs = null;
    System.gc();
  }
  /**
   * Called by constructor.
   * Creates classes stored in classBufs and stores them in classes.
   */
  private void defineClasses() {
    Set s = classBufs.entrySet();
    Iterator i = s.iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      String key = (String)me.getKey();
      if (classes.containsKey(key))
        continue;
      byte[] buf = (byte[])me.getValue();
      try {
        Class jarCl = defineClass(key, buf, 0, buf.length);
        classes.put(key, jarCl);
      }
      catch(NoClassDefFoundError e) {
        handler.log.logprint(toString() + ".loadClassData defineClass(" + key +
          ") failed " + e);
      }
    }
  }
  /**
   * Load classes stored in jarPath archive and stores them in classes.
   * @param  jarPath archive path.
   */
  private void loadClassData(String jarPath) {
    handler.log.logprint(toString() + ".loadClassData(" + jarPath + ")");
    JarInputStream jis = null;
    long lastModified = -1;
    try {
      File f = new File(jarPath);
      lastModified = f.lastModified();
      jis = new JarInputStream(
        new FileInputStream(f));
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".loadClassData reading " + jarPath +
        " " + e);
      return;
    }
    JarEntry je = null;
    String entryName = "";
    try {
      while((je = jis.getNextJarEntry()) != null){
        entryName = je.getName();
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        BufferedInputStream bis = new BufferedInputStream(jis);
        int i;
        while((i = bis.read()) != -1)
          baos.write(i);
        byte[] buf = baos.toByteArray();
        if (entryName.endsWith(".class")) {
          String k = entryName.substring(0,
            entryName.lastIndexOf('.')).replace('/', '.');
          if (classBufs.containsKey(k))
            continue;
          classBufs.put(k, buf);
        } else {
          if (entryName.endsWith("/"))
            continue;
          ResourceEntry re = new ResourceEntry();
          re.buf = buf;
          re.lastModified = lastModified;
          resources.put(entryName, re);
          /* handler.log.logprint(toString() + ".loadClassData resource " +
            entryName); */
        }
        jis.closeEntry();
      }
      jis.close();
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".loadClassData handling " + jarPath +
        "/" + entryName + " " + e);
    }
  }
  /**
   * Called by JSPloader.
   * @param     name name of the class to load.
   * @param     resolve if the class must be resolved.
   * @return    Class
   */
  public synchronized Class loadClass(String name, boolean resolve)
    throws ClassNotFoundException {
    Class cl = (Class)classes.get(name);
    if ((cl == null) && (classBufs != null)) {
      byte[] buf = (byte[])classBufs.get(name);
      if (buf != null) {
        cl = defineClass(name, buf, 0, buf.length);
        classes.put(name, cl);
      }
    }
    handler.log.logprint(toString() + ".loadClass(" + name + ") " +
      (cl == null? "not " : " ") + "found in classes");
    if (cl == null) {
      try {
        cl = findSystemClass(name);
      }
      catch(Exception e) {}
      if (cl != null) {
        /* handler.log.logprint(toString() + ".loadClass(" + name +
          ") found in system classloader"); */
        return cl;
      }
      final String classpath = name;
      cl = (Class)AccessController.doPrivileged(new PrivilegedAction() {
        public Object run() {
          try {
            return handler.servletClassLoader.loadClass(classpath);
          }
          catch(ClassNotFoundException cnfe) {}
          if (parent == null)
            return null;
          try {
            return parent.loadClass(classpath);
          }
          catch(ClassNotFoundException cnfe) {
            return null;
          }
        }
      });
      if (cl == null)
        throw new ClassNotFoundException();
      /*
      else
        handler.log.logprint(toString() + ".loadClass(" + name +
          ") found in other classloaders");
       */
    }
    if (resolve)
      resolveClass(cl);
    return cl;
  }
  /**
   * Externally called method. Also called by JSPloader. In the latter case,
   * it is normal to not find a resource. Note that it is not called in
   * privileged code. So it is better avoiding logging.
   * @param     name name of the resource to stream.
   * @return    InputStream
   */
  public synchronized InputStream getResourceAsStream(String name) {
    ResourceEntry re = (ResourceEntry)resources.get(name);
    if ((re == null) || (re.buf == null)) {
      /* handler.log.errorPrint(toString() + ".getResourceAsStream(" + name +
        ") not found"); */
      return null;
    }
    if (re.buf.length == 0) {
      /* handler.log.errorPrint(toString() + ".getResourceAsStream(" + name +
        ") empty resource"); */
      return null;
    }
    InputStream is = new ByteArrayInputStream(re.buf);
    /*
    handler.log.logprint(toString() + ".getResourceAsStream(" + name +
      (is == null ? ") failed" : ") OK")); */
    return is;
  }
  /**
   * Returns an URL served by JSPservlet.
   * @param name resource name.
   */
  public URL getResource(String name) {
    if (!resources.containsKey(name)) {
      handler.log.errorPrint(toString() + ".getResource(" + name +
        ") not found");
      return null;
    }
    String url = URI + "/" + name;
    handler.log.logprint(toString() + ".getResource(" + name + ") " + url);
    try {
      return new URL(url);
    }
    catch(MalformedURLException mue) {
      return null;
    }
  }
  /**
   * Returns resource archive last modified.
   * @param name resource name.
   */
  long getLastModified(String name) {
    ResourceEntry re = (ResourceEntry)resources.get(name);
    if (re != null)
      return re.lastModified;
    return -1;
  }
  /**
   * Invoked by InputSource in case of Cocoon.
   * Finds the right JSPloader and returns the resource from it.
   * @param     name name of the resource to stream.
   * @return    InputStream
   */
  public synchronized InputStream getResourceAsStreamFromArchive(String name) {
    // Cocoon: In case of xml name contains /, in case of xsl, contains \
    String fullName = name.replace('\\', '/');
    // Depending on the case, the name can start with file:
    if (fullName.startsWith("file:"))
      fullName = fullName.substring("file:".length());
    handler.log.logprint(toString() + ".getResourceAsStreamFromArchive(" +
      fullName + ")");
    // Address InputSource issue... */
    if (fullName.startsWith(URIRealPath)) {
      fullName = fullName.substring(URIRealPath.length());
      /* handler.log.logprint(toString() + ".getResourceAsStreamFromArchive(" +
        fullName + ") without URIRealPath " + URIRealPath); */
      // remove ../
      StringTokenizer st = new StringTokenizer(fullName, "/");
      int nbTokens = st.countTokens();
      if (nbTokens == 0) {
        handler.log.logprint(toString() + ".getResourceAsStreamFromArchive(" +
          fullName + ") no archive name");
        return null;
      }
      String[] tokens = new String[nbTokens];
      int i = 0;
      while(st.hasMoreTokens()) {
        String tok = st.nextToken();
        if (tok.equals("..") && (i > 0))
          -- i;
        else {
          tokens[i] = tok;
          ++ i;
        }
      }
      if (i < 1) {
        handler.log.errorPrint(toString() + ".getResourceAsStreamFromArchive(" +
          fullName + ") " + nbTokens + " tokens found but zero entry");
        return null;
      }
      String newName = "";
      for (int j = 0; j < (i - 1); ++ j)
        newName += (tokens[j] + "/");
      newName += tokens[i - 1];
      fullName = newName;
    }
    /* handler.log.logprint(toString() + ".getResourceAsStreamFromArchive(" +
      fullName + ") without .."); */
    if (fullName.startsWith("/"))
      fullName = fullName.substring(1);
    int idx = fullName.indexOf('/');
    if (idx == -1)
      return null;
    String jarName = fullName.substring(0, idx);
    String classPath = fullName.substring(idx + 1);
    InputStream is = null;
    try {
      is = handler.getResourceAsStream(jarName, classPath);
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() + ".getResourceAsStreamFromArchive(" +
        name + ") " + e);
    }
    handler.log.logprint(toString() + ".getResourceAsStreamFromArchive(" +
      fullName + ") archive:" + jarName + ", path:" + classPath +
      (is == null? " failed" : " OK"));
    return is;
  }
  /**
   * Servlet object table. Must be synchronized -> Hashtable.
   * Key: class name.
   * Value: servlet object.
   */
  private Hashtable cache = new Hashtable();
  /**
   * Returns a servlet instance.
   * @param className servlet class name.
   */
  final Servlet cacheGet(String className) {
    return (Servlet)cache.get(className);
  }
  /**
   * Stores a servlet instance.
   * @param className servlet class name.
   * @param className servlet class name.
   */
  final void cachePut(String className, Servlet servlet) {
    cache.put(className, servlet);
  }
}
