package JSPservletPkg;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import javax.servlet.*;
import java.net.*;
import java.security.*;
/**
 * Class invoked by JSPservlet for a given context path.
 * It forwards the requests to a compiled jsp or servlet loaded by JSPloader.
 * <p>Changes:</p><ul>
 * <li>Version 1.0.1: delete method</li>
 * <li>Version 1.0.3: mapper support</li>
 * <li>Version 1.0.4: mapper persistent cache support, classPath</li></ul>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 * @see JSPservlet
 * @see JSPhandler.ClassEntry
 * @see JSPloader
 */
public class JSPhandler {
  /** path were jars are locally stored - initialized by constructor.
   * @see JSPhandler.ClassEntry
   * @see JSPloader
   */
  String cachePath;
  /** ClassEntry HashMap.
   * key: jar name
   * value: ClassEntry
   */
  HashMap classEntries = new HashMap();
  /** remote locations properties file name - initialized by constructor.
   * @see JSPhandler.ClassEntry
   */
  String remoteLocFile;
  /** Version 1.0.4. Persistent mapper cache file. */
  String mapperCacheFile;
  /** Version 1.0.4. Mapper cache. */
  CacheIntf cache;
  /**
   * Class wraping trace use.
   */
  public class Log {
    /** SimpleDateFormat used to build log entries */
    private final java.text.SimpleDateFormat sdf =
      new java.text.SimpleDateFormat("hh:mm:ss:SSS");
    /** boolean initialized by constructor
     * true: log
     * @see JSPhandler.Log#logprint(String msg)
     */
    private boolean toTrace = false;
    /** PrintWriter initialized by constructor.
     * @see JSPhandler.Log#logprint(String msg)
     */
    private PrintWriter log = null;
    /** log file name */
    String logFile = null;
    /**
     * @param szToTrace initial tracing mode.
     * @param logfile log file path.
     */
    Log(String szToTrace, String logfile) {
      try {
        log = new PrintWriter(new BufferedWriter(new FileWriter(logfile, true)));
      }
      catch(Exception e) {
        log = null;
      }
      if (log != null) {
        logFile = logfile;
        if (szToTrace != null && szToTrace.equalsIgnoreCase("TRUE"))
          toTrace = true;
      }
    }
    /**
     * trace setting method.
     * @param mode new trace mode.
     */
    final synchronized void setTrace(boolean mode) {
      if (log == null)
        return;
      toTrace = mode;
      log.println("INFO  " + sdf.format(new Date()) + " trace set " +
        (mode ? "On":"Off"));
      log.flush();
    }
    /**
     * trace getting method.
     * @return current mode.
     */
    final synchronized boolean getTrace() {
      return toTrace;
    }
    /**
     * user logging method.
     * @param   msg message to format and print.
     */
    final synchronized void userprint(String msg) {
      if (log == null)
        return;
      log.println("USER  " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * logging method.
     * @param   msg message to format and print.
     */
    public final synchronized void logprint(String msg) {
      if (!toTrace)
        return;
      log.println("INFO  " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * Error logging method.
     * @param   msg message to format and print.
     */
    final synchronized void errorPrint(String msg) {
      if (log == null)
        return;
      log.println("ERROR " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /**
     * Exception logging method.
     * @param   msg message to format and print.
     */
    final synchronized void exceptionPrint(String msg) {
      if (log == null)
        return;
      log.println("EXCEP " + sdf.format(new Date()) + " " + msg);
      log.flush();
    }
    /** Empties the log file. */
    final synchronized void clear() {
      if (log == null)
        return;
      log.close();
      try {
        log = new PrintWriter(new BufferedWriter(new FileWriter(logFile, false)));
      }
      catch(Exception e) {
        log = null;
      }
      if (log == null) {
        logFile = null;
        toTrace = false;
      }
    }
    /** to avoid loitering */
    protected void finalize() {
      if (log == null)
        return;
      log.close();
    }
  }
  /** Handle on Log object. */
  public Log log;
  /** remote locations Properties - initialized by constructor.
   * @see JSPhandler#update(String jarName, String remoteLocation)
   * @see JSPhandler.ClassEntry
   */
  Properties remoteLocProp = new Properties();
  /** ServletConfig - provided by JSPservlet.
   * @see JSPhandler.ClassEntry
   */
  ServletConfig servletConfig;
  /** ClassLoader used to create JSPservlet. */
  ClassLoader servletClassLoader;
  /** describes a policy for archive without policy. */
  String defaultPolicy = null;
  /** describes an default policy file and set sandbox for archive execution. */
  String allPermissionPolicy = null;
  /** keystore location. Also states if policy should be downloaded. */
  String keystore = null;
  /** keystore password. */
  String keystorePassword = null;
  /** resources expiration time in seconds - initialized by constructor. */
  int expiration;
  /**
   * URL to access Certificate Revocation List in LDAP.
   * Initialized by constructor.
   */
  String CRLURL = null;
  /**
   * Principal to use to connect to LDAP for CRL.
   * Initialized by constructor.
   */
  String CRLLDAPuser = null;
  /**
   * Password to use to connect to LDAP for CRL.
   * Initialized by constructor.
   */
  String CRLLDAPpasswd = null;
  /**
   * Certificate Revocation List Scan Period in seconds.
   * Initialized by constructor.
   */
  int CRLperiod;
  /**
   * URL to access CA certificate in LDAP.
   * Initialized by constructor.
   */
  String CAURL = null;
  /**
   * Principal to use to connect to LDAP for CA certificate.
   * Initialized by constructor.
   */
  String CALDAPuser = null;
  /**
   * Password to use to connect to LDAP for CA certificate.
   * Initialized by constructor.
   */
  String CALDAPpasswd = null;
  /** record statistics or not */
  boolean toStat = false;
  /** Publisher property file name. */
  String publishers = null;
  /** Version 1.0.3.
   * Publisher Properties - initialized by constructor.<dir>
   * <li>Key: Archive name.</li>
   * <li>Value: Publisher.</li></dir>
   */
  Properties publisherProp = new Properties();
  /** other errors */
  final static int EP_OTHER = 0;
  /** servlet or resource not found */
  final static int EP_NOTFOUND = 1;
  /** servlet attempted an unauthorized access */
  final static int EP_UNAUTHORIZED = 2;
  /** archive certificate revoked */
  final static int EP_REVOKED = 3;
  /** error page array size */
  final static int EP_SIZE = 4;
  /** Scanner wake-up */
  static final int SCANPERIOD = 1000 * 300;
  /** Area of error handling pages */
  byte[][] aErrorPage = new byte[EP_SIZE][];
  /** PageBox unique identifier */
  String ID = null;
  /** Version 1.0.4. Support archives. */
  public String cpath = null;
  /** Version 1.0.4. Support archives. */
  String[] classPath = null;
  /** Version 1.0.4. Class loader for classPath archives. */
  ClassPathLoader classPathLoader;
  /** Version 1.0.4. Pagebox mapping property file name. */
  String pageboxMapping = null;
  /** Version 1.0.4.
   * Suffix mapping - initialized by constructor.
   * <li>Key: Suffix.</li>
   * <li>Value: class name.</li></dir>
   */
  Properties mappingProp = new Properties();
  final byte[] getErrorBuf(ServletContext ctx, String name) {
    String errorPage = servletConfig.getInitParameter(name);
    if (errorPage == null)
      return null;
    try {
      InputStream is = ctx.getResourceAsStream(errorPage);
      if (is == null)
        return null;
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      BufferedInputStream bis = new BufferedInputStream(is);
      int i;
      while((i = bis.read()) != -1)
        baos.write(i);
      return baos.toByteArray();
    }
    catch(Throwable t) {
      return null;
    }
  }
  /**
   * Version 1.0.3: + publishers.<br>
   * Version 1.0.4: + mapperCacheFile, classPath, overClassPath, pageboxMapping.
   * <p>Constructor.
   * <p><ol><li>Initializes parameters from web.xml.</li>
   * <li>Opens the logfile.</li>
   * <li>Restore remoteLocation properties.</li></ol><p>Parameters:
   * <ul>
   * <li>cachePath: path where jar files are are locally stored (permanent cache).
   * </li><li>toTrace: if TRUE, trace messages are stored in logfile.</li>
   * <li>logfile: trace file.</li>
   * <li>remoteLocations: property file where remote jar locations are maintained.
   * </li></ul><p>
   * @param   sc ServletConfig used to initialize target classes.
   * @param   contextPath application prefix.
   * @param   ctx servlet context.
   * @param   URI Request URL up to context path.
   *          Used by ClassPathLoader to handle resources.
   * @param   URIRealPath path without servlet part.
   *          Used by ClassPathLoader to handle resources.
   * @exception JSPloaderException
   *            creation of log failure.
   */
  JSPhandler(ServletConfig sc, String contextPath, ServletContext ctx,
    String URI, String URIRealPath) throws JSPloaderException {
    servletConfig = sc;
    servletClassLoader = JSPservlet.class.getClassLoader();
    cachePath = sc.getInitParameter("cachePath");
    if (cachePath == null)
      cachePath = "C:/temp";
    String szToTrace = sc.getInitParameter("toTrace");
    String logfile = sc.getInitParameter("logfile");
    if (logfile == null)
      logfile = cachePath + "/log.txt";
    log = new Log(szToTrace, logfile);
    String szToStat = sc.getInitParameter("toStat");
    if ((szToStat != null) && (szToStat.equalsIgnoreCase("TRUE")))
      toStat = true;
    ID = sc.getInitParameter("ID");
    remoteLocFile = sc.getInitParameter("remoteLocations");
    if (remoteLocFile == null)
      remoteLocFile = cachePath + contextPath + ".properties";
    mapperCacheFile = sc.getInitParameter("MapperCache");
    if (mapperCacheFile == null)
      mapperCacheFile = cachePath + contextPath + "-mapper.ser";
    cache = new Cache(mapperCacheFile);
    defaultPolicy = sc.getInitParameter("defaultPolicy");
    allPermissionPolicy = sc.getInitParameter("allPermissionPolicy");
    String strExpiration = sc.getInitParameter("expiration");
    if ((strExpiration == null) ||
      ((expiration = Integer.parseInt(strExpiration)) == 0))
      expiration = 5;
    keystore = sc.getInitParameter("keystore");
    keystorePassword = sc.getInitParameter("keystorePassword");
    CRLURL = sc.getInitParameter("CRLURL");
    CAURL = sc.getInitParameter("CAURL");
    CRLLDAPuser = sc.getInitParameter("CRLLDAPuser");
    CRLLDAPpasswd = sc.getInitParameter("CRLLDAPpasswd");
    CALDAPuser = sc.getInitParameter("CALDAPuser");
    CALDAPpasswd = sc.getInitParameter("CALDAPpasswd");
    publishers =  sc.getInitParameter("Publishers");
    if (publishers == null)
      publishers = cachePath + contextPath + "-publisher.properties";
    aErrorPage[EP_OTHER] = getErrorBuf(ctx, "ErrorPage");
    aErrorPage[EP_NOTFOUND] = getErrorBuf(ctx, "NotFoundPage");
    aErrorPage[EP_UNAUTHORIZED] = getErrorBuf(ctx, "UnauthorizedPage");
    aErrorPage[EP_REVOKED] = getErrorBuf(ctx, "RevokedPage");
    String strCRLperiod = sc.getInitParameter("CRLperiod");
    if ((strCRLperiod == null) ||
      ((CRLperiod = Integer.parseInt(strCRLperiod)) == 0))
      CRLperiod = 7 * 24 * 3600;
    cpath = sc.getInitParameter("ClassPath");
    if (cpath != null) {
      StringTokenizer st = new StringTokenizer(cpath, File.pathSeparator);
      classPath = new String[st.countTokens()];
      for(int i = 0; st.hasMoreTokens(); ++i)
        classPath[i] = st.nextToken();
      classPathLoader = new ClassPathLoader(this, URI, URIRealPath);
    }
    pageboxMapping =  sc.getInitParameter("PageboxMapping");
    File f = new File(remoteLocFile);
    if ((f != null) && f.exists()) {
      try {
        remoteLocProp.load(new DataInputStream(new FileInputStream(f)));
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".JSPhandler reading properties:" +
          remoteLocFile + " " + e);
      }
    }
    f = new File(publishers);
    if ((f != null) && f.exists()) {
      try {
        publisherProp.load(new DataInputStream(new FileInputStream(f)));
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".JSPhandler reading properties:" +
          publishers + " " + e);
      }
    }
    if ((pageboxMapping != null) && (pageboxMapping.length() != 0)) {
      f = new File(pageboxMapping);
      if ((f != null) && f.exists()) {
        try {
          mappingProp.load(new DataInputStream(new FileInputStream(f)));
        }
        catch(Exception e) {
          log.errorPrint(toString() + ".JSPhandler reading properties:" +
            pageboxMapping + " " + e);
        }
      }
    }
    new Scanner(this);
    log.logprint(toString() + ".JSPhandler cache path:" + cachePath +
      " instantiated");
  }
  /**
   * Implements a scanner thread.
   * <ul><li>Statistics flush</li>
   * <li>Version 1.0.3: check subscribers</li></ul>
   */
  class Scanner extends Thread {
    /** JSPhandler back pointer */
    JSPhandler handler;
    /** Constructor. Creates the thread.*/
    Scanner(JSPhandler jh) {
      handler = jh;
      start();
    }
    /** Loops to flush Stat entries. */
    public void run() {
      int counterForPing = 0;
      while(true) {
        int remaining = SCANPERIOD;
        long nextTime = System.currentTimeMillis() + remaining;
        while(remaining > 0) {
          try {
            sleep(remaining);
          }
          catch(InterruptedException e) {}
          remaining = (int)(nextTime - System.currentTimeMillis());
        }
        handler.cache.save();
        HashMap hm = null;
        synchronized(handler) {
          hm = (HashMap)handler.classEntries.clone();
        }
        if (++ counterForPing == Integer.MAX_VALUE)
          counterForPing = 0;
        Iterator i = hm.values().iterator();
        while(i.hasNext()) {
          ClassEntry ce = (ClassEntry)i.next();
          ce.stat.flush();
          ce.backCheck.getMapper().checkAll(counterForPing);
        }
      }
    }
  }
  /**
   * Class invoked by JSPhandler for a given jar name.
   * It caches created target servlets and JSPs and manages a JSPloader.
   */
  class ClassEntry {
    /** ClassLoader */
    JSPloader jl;
    /** jar name */
    String jarName;
    /** jar remote location (URL) */
    String jarURL;
    /** HashMap
     * key: class name (full path)
     * value: ServletInfo
     * @see JSPhandler.ClassEntry.ServletInfo
     */
    HashMap servletObjects;
    /** back pointer on JSPhandler */
    JSPhandler handler;
    /**
     * Class of servletObjects entries.
     * It wraps Servlet with its protection domain.
     */
    class ServletInfo {
      /** Embedded servlet. */
      Servlet servlet;
      /** Production Domain wrapper around the PD used to define the class. */
      JSPloader.ProtectionDomainInfo pdi = null;
      /** Version 1.0.4: support of archives without classes.<br>
       * Constructor.
       * @param   ce ClassEntry creating ServletInfo.
       * @param   className class name of the embedded Servlet.
       * @param   servlet embedded Servlet.
       */
      ServletInfo(ClassEntry ce, String className, Servlet servlet) {
        this.servlet = servlet;
        if (ce.jl.classes == null)
          return;
        JSPloader.ClassInfo ci =
          (JSPloader.ClassInfo)ce.jl.classes.get(className);
        if (ci != null)
          pdi = ci.pdi;
      }
      /**
       * check if the credential the class has been defined with are not revoked.
       */
      final boolean check() {
        if (pdi == null)
          return true;
        return pdi.check();
      }
    }
    /**
     * wraps statistic data.
     * uses a Properties to store and persist stats.
     */
    class Stat {
      /** statistic properties */
      Properties prop = null;
      /** ClassEntry back pointer */
      ClassEntry ce = null;
      /** stat file path */
      String statFile = null;
      /**
       * Constructor.
       * @param   ce ClassEntry creating Stat.
       */
      Stat(ClassEntry ce) {
        this.ce = ce;
      }
      /**
       * Statistic initialization.
       * Reads stat property file in memory.
       */
      final synchronized void init() {
        prop = new Properties();
        statFile = ce.handler.cachePath + "/" + ce.jarName + ".stat";
        File f = new File(statFile);
        if ((f != null) && f.exists()) {
          try {
            prop.load(new DataInputStream(new FileInputStream(f)));
          }
          catch(Exception e) {
            ce.handler.log.errorPrint(toString() +
              ".init reading properties:" + statFile + " " + e);
          }
        }
        ce.handler.log.logprint(toString() + ".init for " + statFile);
      }
      /**
       * Statistic increment.
       */
      final synchronized void inc(String classpath) {
        if (prop == null)
          return;
        String szCounter = prop.getProperty(classpath);
        if (szCounter == null)
          szCounter = "1";
        else
          szCounter = new Long(Long.parseLong(szCounter) + 1).toString();
        prop.setProperty(classpath, szCounter);
      }
      /**
       * Stores statistics in property file.
       */
      final synchronized void flush() {
        if ((statFile == null) || (prop == null))
          return;
        try {
          BufferedOutputStream bos = new BufferedOutputStream(
            new FileOutputStream(statFile));
          prop.store(bos, null);
          bos.close();
        }
        catch(Exception e) {
          handler.log.errorPrint(toString() + ".flush writing properties:" +
            statFile + " " + e);
        }
      }
      /**
       * Close property file.
       */
      final synchronized void close() {
        if (prop == null)
          return;
        handler.log.logprint(toString() + ".close for " + statFile);
        flush();
        prop.clear();
        prop = null;
      }
      /** to avoid loitering */
      protected void finalize() {
        if (prop == null)
          return;
        close();
      }
    }
    /** Handle on Stat object. */
    Stat stat = null;
    /** Version 1.0.3. BackCheck handle. Wraps a Mapper. */
    BackCheck backCheck = null;
    /** Version 1.0.4. Mapper persistent cache support.<p>
     * Constructor.
     * <ul><li>get JAR remote location from remoteLocations.</li>
     * <li>Creates JSPloader.</li>
     * <li>Instanciates a servletObjects target cache.</li></ul><p>
     * @param   jh invoking JSPhandler.
     * @param   jarName application prefix.
     * @param   jarURL jar remote location.
     * @exception JSPloaderException
     *            remote URL of jar not defined.
     */
    ClassEntry(JSPhandler jh, String jarName, String jarURL)
      throws JSPloaderException {
      this.jarName = jarName;
      handler = jh;
      this.jarURL = jarURL;
      jl = new JSPloader(jh, jarName, jarURL);
      servletObjects = new HashMap(jl.getSize());
      stat = new Stat(this);
      if (handler.toStat)
        stat.init();
      String publisher = jh.publisherProp.getProperty(jarName);
      if (publisher != null) {
        backCheck = new BackCheck(publisher, jh);
        MapperIntf mapper = backCheck.getMapper();
        String msg = mapper.getMessage();
        if (msg != null)
          jh.log.errorPrint(toString() + ".ClassEntry " + msg);
      }
      handler.log.logprint(toString() + ".ClassEntry(" + jarName + ", " +
        jarURL + ") instantiated");
    }
    /**
     * Clears itself and class loaded stuff.
     */
    final void delete() {
      Collection c = servletObjects.values();
      Iterator i = c.iterator();
      ServletInfo s;
      while(i.hasNext()) {
        s = (ServletInfo)i.next();
        s.servlet.destroy();
      }
      c = null;
      i = null;
      jl = null;
      servletObjects.clear();
      servletObjects = null;
      System.gc();
    }
    /**
     * Clears itself and class loaded stuff and recreates class loader.
     * <ol>
     * <li>Enumerates servletObjects target cache and destroys each target found
     * </li><li>Clear references.</li>
     * <li>Invokes garbage collector.</li>
     * <li>Creates new JSPloader.</li>
     * <li>Instanciates a new servletObjects target cache.</li></ol><p>
     * @param     remoteLocation URL where to download the jar.
     * @exception JSPloaderException
     *            remote URL of jar not defined.
     * @see JSPhandler.ClassEntry.ServletInfo
     */
    final void update(String remoteLocation) throws JSPloaderException {
      if ((remoteLocation != null) && (remoteLocation.length() > 0))
        jarURL = remoteLocation;
      delete();
      jl = new JSPloader(handler, jarName, jarURL);
      servletObjects = new HashMap(jl.getSize());
    }
    /**
     * Given a target classpath, returns a Servlet object.
     * <ol><li>Get the Class corresponding to the target classPath.</li>
     * <li>Instanciate an object of this class and initialize it.</li></ol>
     * @param   classPath full name of the JSP or servlet to load and instantiate.
     * @return    the requested Servlet.
     * @exception JSPloaderException
     *            class loading or instantiation failure.
     * @see JSPhandler.ClassEntry.ServletInfo
     */
    final Servlet get(String classPath) throws JSPloaderException,
      ClassNotFoundException {
      if (servletObjects.containsKey(classPath)) {
        ServletInfo si = (ServletInfo)servletObjects.get(classPath);
        if (!si.check())
          throw new JSPloaderException(handler, toString() + ".get(" + classPath
            + ")", JSPloaderException.JLE_REVOKED);
        stat.inc(classPath);
        return si.servlet;
      }
      Servlet srv = null;
      try {
        String className = classPath.replace('/', '.');
        Class jspClass = jl.loadClass(className);
        if (jspClass == null)
          throw new JSPloaderException(handler, toString() + ".get(" + classPath
            + ") null class");
        srv = (Servlet)jspClass.newInstance();
        if (srv == null)
          throw new JSPloaderException(handler, toString() + ".get(" + classPath
            + ") null object");
        handler.log.logprint(toString() + ".get(" + classPath +
          ") object created");
        srv.init(handler.servletConfig);
        servletObjects.put(classPath, new ServletInfo(this, className, srv));
      }
      catch(JSPloaderException jle) {
        throw jle;
      }
      catch(ClassNotFoundException cne) {
        if (cne.getMessage().indexOf("Invalidated certificate for") != -1) {
          throw new JSPloaderException(handler, toString() + ".get(" + classPath
            + ")", JSPloaderException.JLE_REVOKED);
        } else {
          handler.log.errorPrint(toString() + ".get(" + classPath +
            ") Servlet not found " + cne);
          throw cne;
        }
      }
      catch(Exception e) {
        throw new JSPloaderException(handler, toString() + ".get(" + classPath +
          ") " + e);
      }
      stat.inc(classPath);
      handler.log.logprint(toString() + ".get(" + classPath +
        ") Servlet instantiated");
      return srv;
    }
    /**
     * Invoked by JSPhandler.
     * @param     classPath path to the resource.
     * @return    InputStream on resource.
     */
    final InputStream getResourceAsStream(String classPath) {
      return jl.getResourceAsStream(classPath);
    }
    /**
     * Returns resource timestamp
     * @param     name name of the resource.
     * @return long either<ol>
     * <li>-1: no information (resource managed by container)</li>
     * <li>0: unknown timestamp</li>
     * <li>>1: valid timestamp</li></ol>
     */
    final long getLastModified(String name) {
      long lastModified = jl.getLastModified(name);
      handler.log.logprint(toString() + ".getLastModified resource(" + name +
        ") " + lastModified);
      return lastModified;
    }
    /** Version 1.0.3
     * <p>Returns comment associated with routing if the resource is a routing
     * one or null.</p>
     * @param     name name of the resource to stream.
     */
    final String getRoutingResource(String name) {
      return jl.getRoutingResource(name);
    }
    /** to avoid loitering */
    protected void finalize() {
      stat = null;
      servletObjects.clear();
      servletObjects = null;
      jl = null;
    }
  }
  /**
   * Invoked by JSPservlet.
   * <ol><li>Search ClassEntry with jarName. Creates it if not found</li>
   * <li>invokes ClassEntry.get()</li></ol>
   * @param     jarName first part of the path requested by the browser.
   * @param     classPath last part of the path requested by the browser.
   * @return    the requested Servlet.
   * @exception JSPloaderException see log for reason.
   * @see JSPhandler.ClassEntry#get(String classPath)
   * @see JSPservlet#service(HttpServletRequest request,
   * HttpServletResponse response)
   */
  final synchronized Servlet get(String jarName, String classPath)
    throws JSPloaderException, ClassNotFoundException {
    log.logprint(toString() + ".get(" + jarName + ", " + classPath + ")");
    ClassEntry ce = null;
    if (classEntries.containsKey(jarName)) {
      ce = (ClassEntry)classEntries.get(jarName);
      return ce.get(classPath);
    }
    String jarURL = (String)remoteLocProp.getProperty(jarName);
    if (jarURL == null)
      throw new JSPloaderException(this, toString() + ".get(" + jarName + ")",
        JSPloaderException.JLE_UNDEF);
    ce = new ClassEntry(this, jarName, jarURL);
    classEntries.put(jarName, ce);
    return ce.get(classPath);
  }
  /**
   * Invoked by ServletUpdate.
   * @param     jarName name of the jar.
   */
  public final synchronized void delete(String jarName) {
    log.logprint(toString() + ".delete(" + jarName + ")");
    try {
      File f = new File(cachePath + "/" + jarName + ".jar");
      if ((f != null) && (f.exists()))
        f.delete();
      f = new File(cachePath + "/" + jarName + ".war");
      if ((f != null) && (f.exists()))
        f.delete();
      f = new File(cachePath + "/" + jarName + ".policy");
      if ((f != null) && (f.exists()))
        f.delete();
    }
    catch(Exception e) {
      log.errorPrint(toString() + ".delete(" + jarName + ") " + e);
    }
    ClassEntry ce = null;
    if (classEntries.containsKey(jarName)) {
      ce = (ClassEntry)classEntries.get(jarName);
      ce.delete();
      classEntries.remove(jarName);
    }
    if (remoteLocProp.containsKey(jarName)) {
      remoteLocProp.remove(jarName);
      try {
        BufferedOutputStream bos = new BufferedOutputStream(
          new FileOutputStream(remoteLocFile));
        remoteLocProp.store(bos, null);
        bos.close();
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".delete(" + jarName + ") " + e);
      }
    }
  }
  /** Version 1.0.3.<ul>
   * <li>Invoked by JSPservlet and ServletUpdate.</li>
   * <li>Updates and persists remote properties.</li>
   * <li>Assumes jarName and publisher are not null.<li></ul>
   * @param     jarName name of the jar.
   * @param     remoteLocation URL where to download the jar.
   * @return    boolean false if forbidden operation (archive already published
   *            by another publisher).
   */
  public final synchronized boolean publisherUpdate(String jarName,
    String publisher) throws JSPloaderException {
    log.logprint(toString() + ".publisherUpdate(" + jarName + ", " + publisher +
      ")");
    String oldPublisher = publisherProp.getProperty(jarName);
    if (oldPublisher != null)
      return oldPublisher.equals(publisher);
    publisherProp.setProperty(jarName, publisher);
    try {
      BufferedOutputStream bos = new BufferedOutputStream(
        new FileOutputStream(publishers));
      publisherProp.store(bos, null);
      bos.close();
    }
    catch(Exception e) {
      log.errorPrint(toString() + ".publisherUpdate(" + jarName + ", " +
        publisher + ") " + e);
      return false;
    }
    return true;
  }
  /**
   * Invoked by JSPservlet, ServletUpdate and JSPupdate.
   * <ol><li>Remove local cache</li>
   * <li>If remote location was set by JSPupdate,
   * updates and persists remote properties.</li>
   * <li>Search ClassEntry with jarName. Creates it if not found</li></ol><p>
   * @param     jarName name of the jar.
   * @param     remoteLocation URL where to download the jar.
   * @return    boolean true if jarName found or false otherwise.
   * @exception JSPloaderException
   *            in ClassEntry.get or when persisting modified properties.
   * @see JSPhandler.ClassEntry#update(String remoteLocation)
   * @see JSPservlet#service(HttpServletRequest request,
   * HttpServletResponse response)
   */
  public final synchronized boolean update(String jarName, String remoteLocation)
    throws JSPloaderException {
    log.logprint(toString() + ".update(" + jarName + ", " + remoteLocation + ")");
    try {
      File f = new File(cachePath + "/" + jarName + ".jar");
      if ((f != null) && (f.exists()))
        f.delete();
      f = new File(cachePath + "/" + jarName + ".war");
      if ((f != null) && (f.exists()))
        f.delete();
      f = new File(cachePath + "/" + jarName + ".policy");
      if ((f != null) && (f.exists()))
        f.delete();
    }
    catch(Exception e) {
      log.errorPrint(toString() + ".update(" + jarName + ", " + remoteLocation +
        ") " + e);
    }
    ClassEntry ce = null;
    boolean rc = false;
    if (classEntries.containsKey(jarName)) {
      ce = (ClassEntry)classEntries.get(jarName);
      ce.update(remoteLocation);
      rc = true;
    } else {
      if ((remoteLocation == null) || (remoteLocation.length() == 0)) {
        remoteLocation = (String)remoteLocProp.getProperty(jarName);
        if (remoteLocation == null)
          throw new JSPloaderException(this, toString() + ".update(" +
            jarName + ")", JSPloaderException.JLE_UNDEF);
      }
      ce = new ClassEntry(this, jarName, remoteLocation);
      classEntries.put(jarName, ce);
    }
    if ((remoteLocation != null) && (remoteLocation.length() > 0)) {
      remoteLocProp.setProperty(jarName, remoteLocation);
      try {
        BufferedOutputStream bos = new BufferedOutputStream(
          new FileOutputStream(remoteLocFile));
        remoteLocProp.store(bos, null);
        bos.close();
      }
      catch(Exception e) {
        log.errorPrint(toString() + ".update(" + jarName + ", " + remoteLocation
          + ") " + e);
      }
    }
    return rc;
  }
  /**
   * Invoked by JSPservlet.
   * <ol><li>Find remote location in remoteProperties</li>
   * <li>Find resource in cache.</li>
   * <li>If not found, download it and store it in cache</li></ol><p>
   * @param     jarName name of the jar.
   * @param     classPath path to the resource.
   * @return    InputStream on resource.
   * @see JSPservlet#service(HttpServletRequest request,
   * HttpServletResponse response)
   */
  final InputStream getResourceAsStream(String jarName,
    String classPath) throws JSPloaderException, ClassNotFoundException {
    ClassEntry ce = null;
    synchronized(this) {
      if (classEntries.containsKey(jarName))
        ce = (ClassEntry)classEntries.get(jarName);
      else {
        String jarURL = (String)remoteLocProp.getProperty(jarName);
        if (jarURL == null)
          throw new JSPloaderException(this, toString() +
            ".getResourceAsStream (" + jarName + ")",
            JSPloaderException.JLE_UNDEF);
        ce = new ClassEntry(this, jarName, jarURL);
        classEntries.put(jarName, ce);
      }
    }
    return ce.getResourceAsStream(classPath);
  }
  /**
   * Invoked by JSPservlet.
   * <ol><li>Find remote location in remoteProperties</li>
   * <li>Find resource in cache.</li>
   * <li>If not found, download it and store it in cache</li></ol><p>
   * @param     jarName name of the jar.
   * @param     classPath path to the resource.
   * @return    last modified time of resource.
   * @see JSPservlet#service(HttpServletRequest request,
   * HttpServletResponse response)
   */
  final long getLastModified(String jarName,
    String classPath) throws JSPloaderException, ClassNotFoundException {
    ClassEntry ce = null;
    synchronized(this) {
      if (classEntries.containsKey(jarName))
        ce = (ClassEntry)classEntries.get(jarName);
      else {
        String jarURL = (String)remoteLocProp.getProperty(jarName);
        if (jarURL == null)
          throw new JSPloaderException(this, toString() +
            ".getLastModified (" + jarName + ")",
            JSPloaderException.JLE_UNDEF);
        ce = new ClassEntry(this, jarName, jarURL);
        classEntries.put(jarName, ce);
      }
    }
    return ce.getLastModified(classPath);
  }
  /** Version 1.0.3
   * <p>Invoked by JSPservlet.
   * Returns getRoutingPages for an archive.<p>
   * @param     jarName name of the jar.
   * @return    Properties.
   */
  final Properties getRoutingPages(String jarName)
    throws JSPloaderException, ClassNotFoundException {
    ClassEntry ce = null;
    synchronized(this) {
      if (classEntries.containsKey(jarName))
        ce = (ClassEntry)classEntries.get(jarName);
      else {
        String jarURL = (String)remoteLocProp.getProperty(jarName);
        if (jarURL == null)
          throw new JSPloaderException(this, toString() +
            ".getRoutingPages (" + jarName + ")",
            JSPloaderException.JLE_UNDEF);
        ce = new ClassEntry(this, jarName, jarURL);
        classEntries.put(jarName, ce);
      }
    }
    return ce.jl.getRoutingPages();
  }
  /** Version 1.0.3
   * <p>Invoked by JSPservlet.
   * Returns MapperIntf for an archive.<p>
   * @param     jarName name of the jar.
   * @return    MapperIntf.
   */
  final MapperIntf getMapper(String jarName)
    throws JSPloaderException, ClassNotFoundException {
    ClassEntry ce = null;
    synchronized(this) {
      if (classEntries.containsKey(jarName))
        ce = (ClassEntry)classEntries.get(jarName);
      else {
        String jarURL = (String)remoteLocProp.getProperty(jarName);
        if (jarURL == null)
          throw new JSPloaderException(this, toString() +
            ".getMapper (" + jarName + ")",
            JSPloaderException.JLE_UNDEF);
        ce = new ClassEntry(this, jarName, jarURL);
        classEntries.put(jarName, ce);
      }
    }
    if (ce.backCheck != null)
      return ce.backCheck.getMapper();
    return null;
  }
  /** Version 1.0.3
   * <p>Invoked by JSPservlet.
   * Find if a resource is a routing resource.</p>
   * @param     jarName name of the jar.
   * @param     classPath path to the resource.
   * @return    Mapper if routing resource or null.
   */
  final BackCheck getRoutingResource(String jarName, String classPath)
    throws JSPloaderException, ClassNotFoundException {
    log.logprint(toString() + ".getRoutingResource(" + jarName + ", " +
      classPath + ")");
    ClassEntry ce = null;
    synchronized(this) {
      if (classEntries.containsKey(jarName))
        ce = (ClassEntry)classEntries.get(jarName);
      else {
        String jarURL = (String)remoteLocProp.getProperty(jarName);
        if (jarURL == null)
          throw new JSPloaderException(this, toString() +
            ".getRoutingResource (" + jarName + ")",
            JSPloaderException.JLE_UNDEF);
        ce = new ClassEntry(this, jarName, jarURL);
        classEntries.put(jarName, ce);
      }
    }
    if (ce.getRoutingResource(classPath) == null)
      return null;
    else
      return ce.backCheck;
  }
  /** to avoid loitering */
  protected void finalize() {
    classEntries.clear();
    remoteLocProp.clear();
    classEntries = null;
    remoteLocProp = null;
    log = null;
  }
}