package JSPservletPkg;
import java.util.*;
import java.net.*;
import java.io.*;
/**
 * Handles whois requests in a worker thread.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
class whoisHandler extends Thread {
  /**
   * Returned by whois.
   */
  class CodeRange {
    /** Country code (ccTLD). */
    String code;
    /** First address of the range alloted to the operator. */
    long add1;
    /** Last address of the range alloted to the operator. */
    long add2;
    /** Constructor.
     * @param code Country code (ccTLD).
     * @param add1 First address of the range alloted to the operator.
     * @param add2 Last address of the range alloted to the operator.
     */
    CodeRange(String code, long add1, long add2) {
      this.code = code;
      this.add1 = add1;
      this.add2 = add2;
    }
  }
  /** If true, the thread is ready. */
  boolean isStarted = false;
  /** Handle on JSPhandler set in constructor. */
  JSPhandler handler;
  /**
   * IP address in xxx.xxx.xxx.xxx format.
   * Set by whois and handled in thread.
   */
  String address = null;
  /** Object returned by thread. If null whois request has failed. */
  CodeRange codeRange = null;
  /** If true the thread has completed its task. */
  boolean completed = false;
  /** No operation - wait. */
  private static final int NOOP = 0;
  /** Check availability - use of subscrib and rc. */
  private static final int WHOIS = 1;
  /** Equals NOOP or WHOIS. */
  private int action = NOOP;
  /** Starts the thread. */
  whoisHandler(JSPhandler jh) {
    handler = jh;
    start();
  }
  /**
   * Returns the country code and the alloted range of addresses for a address.
   * Submits to worker thread.
   * @param address requestor address.
   */
  CodeRange whois(String address) {
    if (!isStarted)
      Thread.yield();
    if (!isStarted) {
      handler.log.errorPrint(toString() + ".whois thread not started.");
      return null;
    }
    synchronized(this) {
      codeRange = null;
      completed = false;
      this.address = address;
      action = WHOIS;
      notify();
    }
    for (int i = 0; i < 50; ++i) {
      try {
        Thread.sleep(20);
      }
      catch(InterruptedException ie) {}
      if (completed)
        return codeRange;
    }
    synchronized(this) {
      action = NOOP;
      notify();
    }
    handler.log.errorPrint(toString() + ".whois subscriber don't answer");
    return null;
  }
  /** Main method of Thread. */
  public synchronized void run() {
    isStarted = true;
    while(true) {
      try {
        wait();
      }
      catch(InterruptedException ie) {}
      if (action == NOOP)
        continue;
      if (action == WHOIS)
        codeRange = threadWhois(address);
      completed = true;
      action = NOOP;
      address = null;
    }
  }
  /**
   * Returns the country code and the alloted range of addresses for a address.
   * Validated with mapper.testArin.
   * <p>Algorithm</p><ol>
   * <li>Returns PageBox based on cache</li>
   * <li>Query whois.arinc.net</li>
   * <li>If not managed by arinc, query ripe or apnic</li>
   * <li>Parse the response to find the IP range</li></ol>
   * @param address requestor address.
   */
  private CodeRange threadWhois(String address) {
    try {
      Socket whoSock = new Socket("whois.arin.net", 43);
      PrintStream out = new PrintStream(whoSock.getOutputStream());
      BufferedReader in = new BufferedReader(new InputStreamReader(
        whoSock.getInputStream()));
      out.println(address);
      String line;
      if ((line = in.readLine()) != null) {
        if (line.startsWith("European Regional Internet Registry")) {
          whoSock.close();
          return threadWhois2("whois.ripe.net", address);
        }
        if (line.startsWith("Asia Pacific Network Information Center")) {
          whoSock.close();
          return threadWhois2("whois.apnic.net", address);
        }
      }
      /* arin parsing is not easy.
       * We look at fourth line for 3 fields address - address (older entries?)
       * In this case the country is not set. We assume it is US.
       * Otherwise, we look for single field on fourth and fifth line for the
       * country. */
      int fieldNb = -1;
      StringTokenizer st = null;
      in.readLine();
      for (int i = 0;((i < 2) && ((line = in.readLine()) != null)); ++i) {
        st = new StringTokenizer(line);
        fieldNb = st.countTokens();
        if (fieldNb != 3)
          continue;
        String addr1 = st.nextToken();
        long add1 = Mapper.getLongAddress(addr1);
        if (add1 == -1)
          continue;
        st.nextToken();
        String addr2 = st.nextToken();
        long add2 = Mapper.getLongAddress(addr2);
        if (add2 == -1)
          continue;
        return new CodeRange("US", add1, add2);
      }
      if (fieldNb != 1) {
        line = in.readLine();
        if (line != null)
        st = new StringTokenizer(line);
        if (st.countTokens() != 1) {
          handler.log.errorPrint(toString() +
            ".threadWhois failed to parse arin response for " + address);
          return null;
        }
      }
      String code = st.nextToken();
      while((line = in.readLine()) != null) {
        if (line.indexOf("Netblock:") == -1)
          continue;
        st = new StringTokenizer(line);
        long add1 = -1, add2 = -1;
        while(st.hasMoreTokens()) {
          if (add1 == -1)
            add1 = Mapper.getLongAddress(st.nextToken());
          else {
            add2 = Mapper.getLongAddress(st.nextToken());
            if (add2 != -1)
              return new CodeRange(code, add1, add2);
          }
        }
      }
    }
    catch(IOException ioe) {}
    return null;
  }
  /**
   * Returns the country code and the alloted range of addresses for a address.
   * RIPE/APNIC parsing.<br>
   * Validated with mapper.testRipe.
   * <p>Algorithm</p><ol>
   * <li>Query specified authority</li>
   * <li>If not managed by arinc, query ripe or apnic</li>
   * <li>Parse the response to find the IP range</li></ol>
   * @param authority authority to query (RIPE/APNIC supported).
   * @param address requestor address.
   */
  private CodeRange threadWhois2(String authority, String address) {
    try {
      Socket whoSock = new Socket(authority, 43);
      PrintStream out = new PrintStream(whoSock.getOutputStream());
      BufferedReader in = new BufferedReader(new InputStreamReader(
        whoSock.getInputStream()));
      out.println(address);
      String line;
      long add1 = -1, add2 = -1;
      while((line = in.readLine()) != null) {
        String[] PageBoxes = null;
        if (line.startsWith("inetnum:")) {
          StringTokenizer st = new StringTokenizer(line);
          while(st.hasMoreTokens()) {
            if (add1 == -1)
              add1 = Mapper.getLongAddress(st.nextToken());
            else {
              add2 = Mapper.getLongAddress(st.nextToken());
              if (add2 != -1)
                break;
            }
          }
          if ((add1 == -1) || (add2 == -2)) {
            handler.log.errorPrint(toString() +
              ".threadWhois found no address range in " + line + " (" + address
              + ")");
            return null;
          }
        }
        if (line.startsWith("country:")) {
          StringTokenizer st = new StringTokenizer(line);
          st.nextToken();
          String code = st.nextToken();
          return new CodeRange(code, add1, add2);
        }
      }
    }
    catch(IOException ioe) {}
    return null;
  }
}